"""Messages

This module contains actions for sending messages & approval requests to a Slack channel.

!!! tip "Before you Start"

    Ensure you have created a PPA Slack app, & enabled the relevant slack app scopes.

    See the [Integration](../integration/) page for more information

!!! warning "Supplying Sensitive Data"
    You should ___always___ use a PPA Vault integration to provide sensitive data to a plugin action.
"""

import datetime
import shortuuid
import os

from ppa import appliance, task, events
from typing import Union, Dict, Any

from . import exceptions
from ._client import _create_client, raise_with_message
from .records import WaitHookResponse, Channel


SHARED_WORKSPACE = "/workspace/share/"


@_create_client
def send_message(token: str, channel: Channel, message: Union[str, Dict[str, Any]]) -> None:
    """Send a message to a Slack channel.

!!! question "Slack Channel & Rich Messages"

    View the Slack block kit builder [here](https://app.slack.com/block-kit-builder/) for more information on creating rich messages.

!!! question "Minimum Plugin Version: 2.0.1"

??? info "Input"

    - __token:__ The slack app / bot user [token](https://api.slack.com/authentication/token-types#granular_bot)

    - __channel:__ A [Channels](../inputs_outputs/#channel) dictionary

    - __message:__ the message to send to the channel

??? success "Output"

    Nothing is outputted by this action.

??? example "Simple Message"

    ```YAML linenums="1"
    - slack.messages.send_message:
        message: "Hello from PPA!"
      load:
        channel: channel
        token: slack_bot_token_dev
    ```

??? example "Markdown Message"

    ```YAML linenums="1"
    - slack.messages.send_message:
        message:
          blocks:
            - type: section
              text:
                type: mrkdwn
                text: 'Hello from PPA!'
      load:
        channel: single_channel
        token: slack_bot_token_dev
    ```
    """
    if isinstance(message, str):
        token.chat_postMessage(channel=channel["id"], text=message)
    elif isinstance(message, dict):
        try:
            token.chat_postMessage(channel=channel["id"], blocks=message["blocks"])
        except KeyError:
            raise_with_message(exceptions.BadMessage, "The supplied message is not a valid blocks dictionary")
    else:
        raise_with_message(
            exceptions.BadMessage,
            "The supplied message is invalid, please provide either a simple string message or blocks dictionary",
        )


@_create_client
def send_file(token: str, channel: Channel, message: str, file_name: str) -> None:
    """Send a file to a Slack channel.

!!! question "File Considerations"

    - For the possible file types please see [here](https://api.slack.com/types/file#file_types)

    - There is a 1 megabyte file size limit for files uploaded as [snippets](https://slack.com/intl/en-gb/help/articles/204145658-Create-or-paste-code-snippets-in-Slack#:~:text=From%20your%20desktop%2C%20click%20the,snippet%20in%20the%20Content%20field.&text=Click%20Create%20snippet.).

!!! question "Minimum Plugin Version: 2.0.1"

??? info "Input"

    - __token:__ The slack app / bot user [token](https://api.slack.com/authentication/token-types#granular_bot)

    - __channel:__ A [Channels](../inputs_outputs/#channel) dictionary

    - __message:__ A simple message to send, along with the file

    - __file_name:__ The name of the file that has been saved to the workspace (must be supplied within quotation marks "")

??? success "Output"

    Nothing is outputted by this action.

??? example
    - Save a file to the PPA workspace

    - Send the file to a selected slack channel

    ```YAML linenums="1"
    - ppa_tools.files.write_json_file:
        name: example_file.json
      load:
        contents: sample_message

    - slack.messages.send_file:
        message: Please read the example file
        file_name: "example_file.json"
      load:
        channel: ind_channel
        token: slack_bot_token
    ```
    """
    file_string = os.path.join(SHARED_WORKSPACE, file_name)
    try:
        token.files_upload(channels=channel["id"], initial_comment=message, file=file_string)
    except FileNotFoundError:
        raise_with_message(FileNotFoundError, f"The file '{file_name}' was not found in the task workspace")


@_create_client
def send_approval_request(token: str, channel: Channel, message: str) -> WaitHookResponse:
    """Send an approval request to a Slack channel & wait until it is approved/rejected.

Once the request is responded to, a message will be posted to Slack containing the following:

- Whether the request was approved or rejected
- The username of the person who responded to the request

Example screenshots of all the Slack messages can be found under the example.

!!! question "Slack Channel"

    The message will be sent to the channel supplied to the action.

!!! question "Minimum Plugin Version: 2.0.1"

??? info "Input"

    - __token:__ The slack app / bot user [token](https://api.slack.com/authentication/token-types#granular_bot)

    - __channel:__ A [Channels](../inputs_outputs/#channel) dictionary

    - __message:__ A simple message to send, along with the approval request

??? success "Output"

    A dictionary with the following keys:

    - __approved:__ `true` if the request was approved, `false` if it was not

    - __responder:__ the PPA user who submitted the response

??? example

    ```YAML linenums="1"
    - slack.messages.send_approval_request:
        message: Please approve this request
      load:
        channel: approval_channel
        token: slack_bot_token
      save: approval_response
    ```

??? note "Approval Request Screenshot"

    <img src="../../../images/screenshots/slack/approval_request.png" alt="approval-request" width="100%"/>

??? note "Approved Message Screenshot"

    <img src="../../../images/screenshots/slack/request_approved.png" alt="requested-approved" width="100%"/>

??? note "Rejected Message Screenshot"

    <img src="../../../images/screenshots/slack/request_rejected.png" alt="request-rejected" width="100%"/>
    """
    channel_id = channel["id"]

    wait_hook_url, secret = events.create_wait_hook()
    approve_link = f"{wait_hook_url}?approve=true"
    reject_link = f"{wait_hook_url}?approve=false"
    task_link = f"https://{appliance.address()}/ui/activity/attach/?uuid={wait_hook_url.split('/')[-2]}"
    task_name = task.manifest()["labels"]["ppa.name"]
    request_uuid = shortuuid.uuid()

    try:
        started_by = task.user().split("\\")[1]
    except IndexError:
        started_by = task.user()

    request_info = f"""*Request Info*

Task: {task_name}

Operator: {started_by}

Requested At: {datetime.datetime.utcnow().strftime('%H:%M:%S %d/%m/%Y')}

Unique Request ID: {request_uuid}

Click <{task_link}|here> to view the task.
    """

    request_message = f"""*Request Message*

{message}
"""

    request_actions = f"""*Please Choose an Option*


:white_check_mark: <{approve_link}|Approve>


:x: <{reject_link}|Reject>"""

    request_body = {
        "blocks": [
            {"type": "divider"},
            {"type": "section", "text": {"type": "mrkdwn", "text": ":bell: *New PPA Approval Request* :bell:"}},
            {"type": "divider"},
            {"type": "section", "text": {"type": "mrkdwn", "text": request_info}},
            {"type": "divider"},
            {"type": "section", "text": {"type": "mrkdwn", "text": request_message}},
            {"type": "divider"},
            {"type": "section", "text": {"type": "mrkdwn", "text": request_actions}},
            {
                "type": "section",
                "text": {"type": "mrkdwn", "text": "_You will be asked to authenticate to PPA before your choice is submitted._"},
            },
            {"type": "divider"},
        ]
    }

    token.chat_postMessage(channel=channel_id, blocks=request_body["blocks"])

    approval_response = WaitHookResponse(**events.wait("Waiting for Slack approval response...", secret=secret))
    approved = approval_response["values"]["approve"][0].lower() == "true"

    try:
        responder = approval_response["username"].split("\\")[1]
    except IndexError:
        responder = approval_response["username"]

    if not approved:
        approval_response_message = f""":rotating_light: *Request Rejected* :rotating_light:

Rejected By: {responder}

Task: {task_name}

Operator: {started_by}

Unique Request ID: {request_uuid}"""

    else:
        approval_response_message = f""":trophy: *Request Approved* :trophy:

Approved By: {responder}

Task: {task_name}

Operator: {started_by}

Unique Request ID: {request_uuid}"""

    approval_message = {
        "blocks": [
            {"type": "divider"},
            {"type": "section", "text": {"type": "mrkdwn", "text": approval_response_message}},
            {"type": "divider"},
        ]
    }

    token.chat_postMessage(channel=channel_id, blocks=approval_message["blocks"])

    return {"approved": approved, "responder": responder}
